open Syntax
open Interpret
open Priorityqueue
open Minisatinterface

exception NoOption

let trainingdataout = ref None
let prioritysalt : int32 option ref = ref None
let self_matings = ref false
let self_confrontations = ref false
let eqn_sym_rule = ref false
let ordered_shallow_rules = ref false
let ordered_shallow_rules_strict = ref false
let ordered_shallow_rules_salt = ref 0l

exception Timeout
exception ScheduleTimeout

(*** set_timer for timeout signal, see http://alan.petitepomme.net/cwn/2006.01.03.html#2 ***)
let set_timer s =
  ignore (Sys.signal Sys.sigalrm (Sys.Signal_handle (fun signo -> raise Timeout)));
  ignore (Unix.setitimer Unix.ITIMER_REAL { Unix.it_interval = 0.0; Unix.it_value = s });;

let set_schedule_timer s =
  ignore (Sys.signal Sys.sigalrm (Sys.Signal_handle (fun signo -> raise ScheduleTimeout)));
  ignore (Unix.setitimer Unix.ITIMER_REAL { Unix.it_interval = 0.0; Unix.it_value = s });;

exception AbstractTimeout

let abstract_time = ref 0
let abstract_time_limit = ref 10000000
let middle_abstract_time = ref 0
let middle_abstract_time_limit = ref 10000000
let middle_search = ref true
let prop_endphase = ref false

let incr_abstract_time () =
  incr abstract_time;
  if !abstract_time > !abstract_time_limit then raise AbstractTimeout;
  if !middle_search then
    begin
      incr middle_abstract_time;
      if !middle_abstract_time > !middle_abstract_time_limit then middle_search := false
    end

let clauses : int list list ref = ref []
let clause_assumption : (int list,unit) Hashtbl.t = Hashtbl.create 100
let clause_princ : (int list,int list) Hashtbl.t = Hashtbl.create 100
let clause_inst : (int list,tm) Hashtbl.t = Hashtbl.create 100

let rote_strict : bool ref = ref false
let eigen : (string,string) Hashtbl.t = Hashtbl.create 100
let eigensasinsts : bool ref = ref false
let pos_rote_shallow : (string,unit) Hashtbl.t = Hashtbl.create 100
let neg_rote_shallow : (string,unit) Hashtbl.t = Hashtbl.create 100
let pos_rote_pos : (string,unit) Hashtbl.t = Hashtbl.create 100
let pos_rote_neg : (string,unit) Hashtbl.t = Hashtbl.create 100
let pos_rote_inst : (string,unit) Hashtbl.t = Hashtbl.create 100
let neg_rote_pos : (string,unit) Hashtbl.t = Hashtbl.create 100
let neg_rote_neg : (string,unit) Hashtbl.t = Hashtbl.create 100
let neg_rote_inst : (string,unit) Hashtbl.t = Hashtbl.create 100
let pos_watch_shallow : (string,unit) Hashtbl.t = Hashtbl.create 100
let neg_watch_shallow : (string,unit) Hashtbl.t = Hashtbl.create 100
let pos_watch_pos : (string,unit) Hashtbl.t = Hashtbl.create 100
let pos_watch_neg : (string,unit) Hashtbl.t = Hashtbl.create 100
let pos_watch_inst : (string,unit) Hashtbl.t = Hashtbl.create 100
let neg_watch_pos : (string,unit) Hashtbl.t = Hashtbl.create 100
let neg_watch_neg : (string,unit) Hashtbl.t = Hashtbl.create 100
let neg_watch_inst : (string,unit) Hashtbl.t = Hashtbl.create 100
let shallowatom : (int,unit) Hashtbl.t = Hashtbl.create 100
let funcdef : (string,tm) Hashtbl.t = Hashtbl.create 100
let preddef : (string,tm) Hashtbl.t = Hashtbl.create 100
let schemata : (string,tm) Hashtbl.t = Hashtbl.create 100
let tmid_schema : (string,string * string list) Hashtbl.t = Hashtbl.create 100
let pos_src_shallow : (string,string) Hashtbl.t = Hashtbl.create 100
let pos_src_pos : (string,string) Hashtbl.t = Hashtbl.create 100
let pos_src_neg : (string,string) Hashtbl.t = Hashtbl.create 100
let pos_src_inst : (string,string) Hashtbl.t = Hashtbl.create 100
let neg_src_shallow : (string,string) Hashtbl.t = Hashtbl.create 100
let neg_src_pos : (string,string) Hashtbl.t = Hashtbl.create 100
let neg_src_neg : (string,string) Hashtbl.t = Hashtbl.create 100
let neg_src_inst : (string,string) Hashtbl.t = Hashtbl.create 100
let shallow_src_pos : (string,string) Hashtbl.t = Hashtbl.create 100
let shallow_src_neg : (string,string) Hashtbl.t = Hashtbl.create 100
let inst_src_neg : (string,string) Hashtbl.t = Hashtbl.create 100

let apply_funcrules_poseqns = ref false

let log_const_p k = k = !fal || k = !neg || k = !iff || k = !disj || k = !conj || k = !exSet || k = !tru || k = !eqSet || k = !expoly || k = !eqPoly || k = !neqSet || k = !neqPoly

let tm_schema_and_instnames m =
  let instcnt = ref 0 in
  let instnamesh : (string,int) Hashtbl.t = Hashtbl.create 10 in
  let instnames = ref [] in
  let rec tm_schema_and_instnames_r m i =
    match m with
    | TmH(h) when not (log_const_p h) ->
      begin
        try
          let j = Hashtbl.find instnamesh h in
          DB(j+i)
        with Not_found ->
          let j = !instcnt in
          incr instcnt;
          instnames := h::!instnames;
          Hashtbl.add instnamesh h j;
          DB(j+i)
      end
    | TpAp(TmH(h),Set) when h = !expoly -> TmH(!exSet)
    | TpAp(TmH(h),Set) when h = !eqPoly -> TmH(!eqSet)
    | TpAp(TmH(h),Set) when h = !neqPoly -> TmH(!neqSet)
    | Ap(m1,m2) ->
        let m1s = tm_schema_and_instnames_r m1 i in
        let m2s = tm_schema_and_instnames_r m2 i in
        Ap(m1s,m2s)
    | Lam(a,m1) ->
        let m1s = tm_schema_and_instnames_r m1 (i+1) in
        Lam(a,m1s)
    | Imp(m1,m2) ->
        let m1s = tm_schema_and_instnames_r m1 i in
        let m2s = tm_schema_and_instnames_r m2 i in
        Imp(m1s,m2s)
    | All(a,m1) ->
        let m1s = tm_schema_and_instnames_r m1 (i+1) in
        All(a,m1s)
    | _ -> m
  in
  let m = tm_schema_and_instnames_r m 0 in
  (m,List.rev !instnames)

let tm_schema m =
  let (m,_) = tm_schema_and_instnames m in
  m

let schema_id m =
  let h = tm_id m in
  if not (Hashtbl.mem schemata h) then Hashtbl.add schemata h m;
  h

let lin_shallow_func_rules : (int * tm,string * int list * tm * tm list * tm list) Hashtbl.t = Hashtbl.create 100
let lin_shallow_pred_rules : (int * tm * bool,string * int list * tm list * tm list) Hashtbl.t = Hashtbl.create 100

let shallow_func_rules : (string,string * int list * int * int list * tm * tm list * tm list) Hashtbl.t = Hashtbl.create 100
let shallow_pos_pred_rules : (string,string * int list * int * int list * tm list * tm list) Hashtbl.t = Hashtbl.create 100
let shallow_neg_pred_rules : (string,string * int list * int * int list * tm list * tm list) Hashtbl.t = Hashtbl.create 100
let shallow_func_func_rules : (string * int * string,string * int list * int * int list * int list * int list * tm * tm list * tm list) Hashtbl.t = Hashtbl.create 100
let shallow_pos_pred_func_rules : (string * int * string,string * int list * int * int list * int list * int list * tm list * tm list) Hashtbl.t = Hashtbl.create 100
let shallow_neg_pred_func_rules : (string * int * string,string * int list * int * int list * int list * int list * tm list * tm list) Hashtbl.t = Hashtbl.create 100

let invert_negprop p =
  match p with
  | Imp(p1,TmH(k)) when k = !fal -> p1
  | Ap(TmH(k),p1) when k = !neg -> p1
  | _ -> raise Not_found

let negprop p =
  try
    invert_negprop p
  with Not_found -> Imp(p,TmH(!fal))

let neg_split_by_polarity pl =
  let posl = ref [] in
  let negl = ref [] in
  List.iter
    (fun p -> try let q = invert_negprop p in posl := q::!posl with Not_found -> negl := p::!negl)
    pl;
  (!posl,!negl)

let rec bvars_spine sp =
  match sp with
  | (DB(j)::spr) -> j::bvars_spine spr
  | [] -> []
  | _ -> raise Not_found

let rec almost_bvars_spine sp pos =
  match sp with
  | (DB(j)::spr) ->
    let (jl1,jl3,pos,m) = almost_bvars_spine spr (pos+1) in
    (j::jl1,jl3,pos,m)
  | (m::spr) -> let jl3 = bvars_spine spr in ([],jl3,pos,m)
  | _ -> raise Not_found

let complete_bvars_p jl i =
  try
    for k = 0 to i-1 do if not (List.mem k jl) then raise Exit done;
    true
  with Exit -> false

let uncovered_bvars jl i =
  let kl = ref [] in
  for k = i-1 downto 0 do if not (List.mem k jl) then kl := k::!kl done;
  !kl

let use_quant_cases = ref false
let use_quant_cases_pre = ref false
let quant_cases = ref []
      
let use_shallowrules_pred = ref true
let use_shallowrules_func = ref true
let use_shallowrules_pred_onlylinear = ref false
let use_shallowrules_func_onlylinear = ref false
let use_shallowrules_pred_onlyalmostlinear = ref false
let use_shallowrules_func_onlyalmostlinear = ref false
let shallow_var_elim = ref true
let dynamic_shallowrules = ref true

type condtype =
  | Eigen of string
  | AssertedPos of string
  | AssertedNeg of string
  | ProcessedPos of string
  | ProcessedNeg of string
  | ProcessedInst of string

let condtype_to_str ct =
  match ct with
  | Eigen(h) -> Printf.sprintf "eigen %s" h
  | AssertedPos(h) -> Printf.sprintf "assertedpos %s" h
  | AssertedNeg(h) -> Printf.sprintf "assertedneg %s" h
  | ProcessedPos(h) -> Printf.sprintf "processedpos %s" h
  | ProcessedNeg(h) -> Printf.sprintf "processedneg %s" h
  | ProcessedInst(h) -> Printf.sprintf "processedinst %s" h

let condstable : (condtype * string list,unit) Hashtbl.t = Hashtbl.create 10

let cutproductionruleg : (condtype,string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let insttrmproductionruleg : (condtype,string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let goodposproductionruleg : (condtype,string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let badposproductionruleg : (condtype,string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let goodnegproductionruleg : (condtype,string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let badnegproductionruleg : (condtype,string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let goodinstproductionruleg : (condtype,string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let badinstproductionruleg : (condtype,string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let cutproductionrule : (condtype,(string * string) list * string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let insttrmproductionrule : (condtype,(string * string) list * string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let goodposproductionrule : (condtype,(string * string) list * string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let badposproductionrule : (condtype,(string * string) list * string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let goodnegproductionrule : (condtype,(string * string) list * string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let badnegproductionrule : (condtype,(string * string) list * string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let goodinstproductionrule : (condtype,(string * string) list * string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let badinstproductionrule : (condtype,(string * string) list * string list * (condtype * string list) list * string * string list) Hashtbl.t = Hashtbl.create 10
let goodproprule : (bool * string,string list * (condtype * string list) list) Hashtbl.t = Hashtbl.create 10
let badproprule : (bool * string,string list * (condtype * string list) list) Hashtbl.t = Hashtbl.create 10
let goodinstrule : (string,string list * (condtype * string list) list) Hashtbl.t = Hashtbl.create 10
let badinstrule : (string,string list * (condtype * string list) list) Hashtbl.t = Hashtbl.create 10
let alwaysgoodpos : (string,unit) Hashtbl.t = Hashtbl.create 10
let alwaysbadpos : (string,unit) Hashtbl.t = Hashtbl.create 10
let alwaysgoodneg : (string,unit) Hashtbl.t = Hashtbl.create 10
let alwaysbadneg : (string,unit) Hashtbl.t = Hashtbl.create 10
let alwaysgoodinst : (string,unit) Hashtbl.t = Hashtbl.create 10
let alwaysbadinst : (string,unit) Hashtbl.t = Hashtbl.create 10

let using_guidance = ref false
let guidance_cuts = ref []
let guidance_insttrms = ref []
let prop_guidance : (bool * bool * string * string list,unit) Hashtbl.t = Hashtbl.create 10
let inst_guidance : (bool * string * string list,unit) Hashtbl.t = Hashtbl.create 10

exception Incompat

let rec extend_guidance_assign al yl il =
  match (yl,il) with
  | ([],[]) -> al
  | (y::yr,i::ir) ->
     begin
       try
         let b = List.assoc y al in
         if b = i then
           extend_guidance_assign al yr ir
         else
           raise Incompat
       with Not_found ->
         extend_guidance_assign ((y,i)::al) yr ir
     end
  | _ -> raise Incompat

let apply_cut_forward_guidance_rules al il yl conds schid xl =
  try
    let al = extend_guidance_assign al yl il in
    if conds = [] then
      begin
        try
          let sch = Hashtbl.find schemata schid in
          let theta = List.map (fun x -> TmH(List.assoc x al)) xl in
          let p = tmssubst sch theta 0 in
          if !verbosity > 9 then Printf.printf "Guidance produced cut %s\n" (s_tm_to_str p);
          guidance_cuts := p::!guidance_cuts
        with Not_found -> () (** shouldn't happen, but ignore it if it does out of laziness **)
      end
    else
      List.iter
        (fun (ct,yl) ->
          Hashtbl.add cutproductionrule ct (al,yl,List.filter (fun c -> not (c = (ct,yl))) conds,schid,xl))
        conds
  with Incompat -> ()

let apply_insttrm_forward_guidance_rules al il yl conds schid xl =
  try
    let al = extend_guidance_assign al yl il in
    if conds = [] then
      begin
        try
          let sch = Hashtbl.find schemata schid in
          let theta = List.map (fun x -> TmH(List.assoc x al)) xl in
          let m = tmssubst sch theta 0 in
          if !verbosity > 9 then Printf.printf "Guidance produced insttrm %s\n" (s_tm_to_str m);
          guidance_insttrms := m::!guidance_insttrms
        with Not_found -> () (** shouldn't happen, but ignore it if it does out of laziness **)
      end
    else
      List.iter
        (fun (ct,yl) ->
          Hashtbl.add insttrmproductionrule ct (al,yl,List.filter (fun c -> not (c = (ct,yl))) conds,schid,xl))
        conds
  with Incompat -> ()

let apply_prop_forward_guidance_rules ht pos good al il yl conds schid xl =
  try
    let al = extend_guidance_assign al yl il in
    if conds = [] then
      begin
        try
          let jl = List.map (fun x -> List.assoc x al) xl in
          Hashtbl.add prop_guidance (good,pos,schid,jl) ();
        with Not_found -> () (** shouldn't happen, but ignore it if it does out of laziness **)
      end
    else
      List.iter
        (fun (ct,yl) ->
          Hashtbl.add ht ct (al,yl,List.filter (fun c -> not (c = (ct,yl))) conds,schid,xl))
        conds
  with Incompat -> ()

let apply_inst_forward_guidance_rules ht good al il yl conds schid xl =
  try
    let al = extend_guidance_assign al yl il in
    if conds = [] then
      begin
        try
          let jl = List.map (fun x -> List.assoc x al) xl in
          Hashtbl.add inst_guidance (good,schid,jl) ();
        with Not_found -> () (** shouldn't happen, but ignore it if it does out of laziness **)
      end
    else
      List.iter
        (fun (ct,yl) ->
          Hashtbl.add ht ct (al,yl,List.filter (fun c -> not (c = (ct,yl))) conds,schid,xl))
        conds
  with Incompat -> ()

let apply_forward_guidance_rules ct il =
  List.iter
    (fun (yl,conds,schid,xl) -> apply_cut_forward_guidance_rules [] il yl conds schid xl)
    (Hashtbl.find_all cutproductionruleg ct);
  List.iter
    (fun (al,yl,conds,schid,xl) -> apply_cut_forward_guidance_rules al il yl conds schid xl)
    (Hashtbl.find_all cutproductionrule ct);
  List.iter
    (fun (yl,conds,schid,xl) -> apply_insttrm_forward_guidance_rules [] il yl conds schid xl)
    (Hashtbl.find_all insttrmproductionruleg ct);
  List.iter
    (fun (al,yl,conds,schid,xl) -> apply_insttrm_forward_guidance_rules al il yl conds schid xl)
    (Hashtbl.find_all insttrmproductionrule ct);
  List.iter
    (fun (yl,conds,schid,xl) -> apply_prop_forward_guidance_rules goodposproductionrule true true [] il yl conds schid xl)
    (Hashtbl.find_all goodposproductionruleg ct);
  List.iter
    (fun (al,yl,conds,schid,xl) -> apply_prop_forward_guidance_rules goodposproductionrule true true al il yl conds schid xl)
    (Hashtbl.find_all goodposproductionrule ct);
  List.iter
    (fun (yl,conds,schid,xl) -> apply_prop_forward_guidance_rules badposproductionrule true false [] il yl conds schid xl)
    (Hashtbl.find_all badposproductionruleg ct);
  List.iter
    (fun (al,yl,conds,schid,xl) -> apply_prop_forward_guidance_rules badposproductionrule true false al il yl conds schid xl)
    (Hashtbl.find_all badposproductionrule ct);
  List.iter
    (fun (yl,conds,schid,xl) -> apply_prop_forward_guidance_rules goodnegproductionrule false true [] il yl conds schid xl)
    (Hashtbl.find_all goodnegproductionruleg ct);
  List.iter
    (fun (al,yl,conds,schid,xl) -> apply_prop_forward_guidance_rules goodnegproductionrule false true al il yl conds schid xl)
    (Hashtbl.find_all goodnegproductionrule ct);
  List.iter
    (fun (yl,conds,schid,xl) -> apply_prop_forward_guidance_rules badnegproductionrule false false [] il yl conds schid xl)
    (Hashtbl.find_all badnegproductionruleg ct);
  List.iter
    (fun (al,yl,conds,schid,xl) -> apply_prop_forward_guidance_rules badnegproductionrule false false al il yl conds schid xl)
    (Hashtbl.find_all badnegproductionrule ct);
  List.iter
    (fun (yl,conds,schid,xl) -> apply_inst_forward_guidance_rules goodinstproductionrule true [] il yl conds schid xl)
    (Hashtbl.find_all goodinstproductionruleg ct);
  List.iter
    (fun (al,yl,conds,schid,xl) -> apply_inst_forward_guidance_rules goodinstproductionrule true al il yl conds schid xl)
    (Hashtbl.find_all goodinstproductionrule ct);
  List.iter
    (fun (yl,conds,schid,xl) -> apply_inst_forward_guidance_rules badinstproductionrule false [] il yl conds schid xl)
    (Hashtbl.find_all badinstproductionruleg ct);
  List.iter
    (fun (al,yl,conds,schid,xl) -> apply_inst_forward_guidance_rules badinstproductionrule false al il yl conds schid xl)
    (Hashtbl.find_all badinstproductionrule ct)

let pos_guidance_prop_3 b p sch il schid =
  if (Hashtbl.mem (if b then alwaysgoodpos else alwaysgoodneg) schid || Hashtbl.mem prop_guidance (true,b,schid,il)) then
    true
  else
    begin
      try
        List.iter
          (fun (xl,conds) ->
            let al = List.map2 (fun x a -> (x,a)) xl il in
            try
              ignore
                (List.find
                   (fun (ct,yl) ->
                     let jl = List.map (fun y -> try List.assoc y al with Not_found -> "fake") yl in (* fake shouldn't happen, but in case it does it won't be in the condstable anyway *)
                     not (Hashtbl.mem condstable (ct,jl)))
                   conds)
            with Not_found -> raise Exit)
          (Hashtbl.find_all goodproprule (b,schid));
        false
      with Exit -> true
    end
  
let pos_guidance_prop_2 b p sch il =
  let schid = tm_id sch in
  pos_guidance_prop_3 b p sch il schid

let pos_guidance_prop b p =
  let (sch,il) = tm_schema_and_instnames p in
  pos_guidance_prop_2 b p sch il
  
let neg_guidance_prop_3 b p sch il schid =
  if (Hashtbl.mem (if b then alwaysbadpos else alwaysbadneg) schid || Hashtbl.mem prop_guidance (false,b,schid,il)) then
    true
  else
    begin
      try
        List.iter
          (fun (xl,conds) ->
            let al = List.map2 (fun x a -> (x,a)) xl il in
            try
              ignore
                (List.find
                   (fun (ct,yl) ->
                     let jl = List.map (fun y -> try List.assoc y al with Not_found -> "fake") yl in (* fake shouldn't happen, but in case it does it won't be in the condstable anyway *)
                     not (Hashtbl.mem condstable (ct,jl)))
                   conds)
            with Not_found -> raise Exit)
          (Hashtbl.find_all badproprule (b,schid));
        false
      with Exit -> true
    end

let neg_guidance_prop_2 b p sch il =
  let schid = tm_id sch in
  neg_guidance_prop_3 b p sch il schid
    
let neg_guidance_prop b p =
  let (sch,il) = tm_schema_and_instnames p in
  neg_guidance_prop_2 b p sch il

let pos_guidance_inst_3 m sch il schid =
  if (Hashtbl.mem alwaysgoodinst schid || Hashtbl.mem inst_guidance (true,schid,il)) then
    true
  else
    begin
      try
        List.iter
          (fun (xl,conds) ->
            let al = List.map2 (fun x a -> (x,a)) xl il in
            try
              ignore
                (List.find
                   (fun (ct,yl) ->
                     let jl = List.map (fun y -> List.assoc y al) yl in
                     not (Hashtbl.mem condstable (ct,jl)))
                   conds)
            with Not_found -> raise Exit)
          (Hashtbl.find_all goodinstrule schid);
        false
      with Exit -> true
    end

let pos_guidance_inst_2 m sch il =
  let schid = tm_id sch in
  pos_guidance_inst_3 m sch il schid

let pos_guidance_inst m =
  let (sch,il) = tm_schema_and_instnames m in
  pos_guidance_inst_2 m sch il

let neg_guidance_inst_3 m sch il schid =
  if (Hashtbl.mem alwaysbadinst schid || Hashtbl.mem inst_guidance (false,schid,il)) then
    true
  else
    begin
      try
        List.iter
          (fun (xl,conds) ->
            let al = List.map2 (fun x a -> (x,a)) xl il in
            try
              ignore
                (List.find
                   (fun (ct,yl) ->
                     let jl = List.map (fun y -> List.assoc y al) yl in
                     not (Hashtbl.mem condstable (ct,jl)))
                   conds)
            with Not_found -> raise Exit)
          (Hashtbl.find_all badinstrule schid);
        false
      with Exit -> true
    end

let neg_guidance_inst_2 m sch il =
  let schid = tm_id sch in
  neg_guidance_inst_3 m sch il schid

let neg_guidance_inst m =
  let (sch,il) = tm_schema_and_instnames m in
  neg_guidance_inst_2 m sch il

type shallowrule =
  | ShallowFunc of int list * int * tm * tm * tm list
  | ShallowPred of int list * int * tm * bool * tm list

let rec allset_close i p =
  if i > 0 then
    allset_close (i-1) (All(Set,p))
  else
    p

let srule_prop srule =
  match srule with
  | ShallowFunc(al,j,m1,m2,othl) ->
      let concl = Ap(Ap(TmH(!eqSet),m1),m2) in
      let p1 = List.fold_right (fun q p -> Imp(q,p)) othl concl in
      allset_close j p1
  | ShallowPred(al,j,p,b,othl) ->
      let concl = if b then negprop p else p in
      let p1 = List.fold_right (fun q p -> Imp(q,p)) othl concl in
      allset_close j p1
	
let pinv_subst i theta =
  let a = Array.make i (TmH("unset")) in
  let presideeqns = ref [] in
  let rec pinv_subst_r theta k =
    match theta with
    | DB(j)::thetar ->
	if a.(j) = TmH("unset") then
	  a.(j) <- DB(k)
	else
	  presideeqns := (k,DB(j))::!presideeqns;
	pinv_subst_r thetar (k+1)
    | m::thetar ->
	presideeqns := (k,m)::!presideeqns;
	pinv_subst_r thetar (k+1)
    | [] -> ()
  in
  pinv_subst_r theta 0;
  let thetai = Array.to_list a in
  (thetai,List.map (fun (k,m) -> Ap(Ap(TmH(!eqSet),DB(k)),tmssubst2 m thetai 0)) !presideeqns)

let shallow_subst i jl s =
  let a = Array.make i (TmH("unset")) in
  let sideeqns = ref [] in
  let rec shallow_subst_r jl s =
    match (jl,s) with
    | (j::jr,arg::sr) ->
      if a.(j) = TmH("unset") then
        a.(j) <- arg
      else
        sideeqns := (Ap(Ap(TmH(!eqSet),arg),a.(j)))::!sideeqns;
      shallow_subst_r jr sr
    | ([],[]) -> ()
    | _ -> raise (Failure "arity mismatch in shallow_subst")
  in
  shallow_subst_r jl s;
  (Array.to_list a,!sideeqns)

let ordered_salted_h : (string,int32) Hashtbl.t = Hashtbl.create 100

let ordered_salted_v h =
    try
      Hashtbl.find ordered_salted_h h
    with Not_found ->
      let (v,_,_,_,_,_,_,_) = Hash.hashtag (Hash.hexstring_hashval h) !ordered_shallow_rules_salt in
      Hashtbl.add ordered_salted_h h v;
      if !verbosity > 9 then Printf.printf "ordered_salted_v %s = %ld\n" h v;
      v

let rec ensure_all_instnames_smaller m vbd =
    match m with
    | TmH(h) when not (log_const_p h) ->
      let v = ordered_salted_v h in
      if v > vbd || (!ordered_shallow_rules_strict && v = vbd) then
        begin
          if !verbosity > 9 then Printf.printf "instname %s has val %ld exceeding bd %ld\n" h v vbd;
	  raise Exit
        end
    | Ap(m1,m2) -> ensure_all_instnames_smaller m1 vbd; ensure_all_instnames_smaller m2 vbd
    | Lam(_,m1) -> ensure_all_instnames_smaller m1 vbd
    | Imp(m1,m2) -> ensure_all_instnames_smaller m1 vbd; ensure_all_instnames_smaller m2 vbd
    | All(_,m1) -> ensure_all_instnames_smaller m1 vbd
    | _ -> ()

let check_ordered_shallow_rule_ok srule =
  match srule with
  | ShallowFunc(al,j,m1,m2,othl) ->
    let (_,insl) = tm_schema_and_instnames m1 in
    let maxv = ref Int32.min_int in
    List.iter (fun h -> let v = ordered_salted_v h in if v > !maxv then maxv := v) insl;
    begin
      try
        ensure_all_instnames_smaller m2 !maxv;
        List.iter (fun q -> ensure_all_instnames_smaller q !maxv) othl
      with Exit ->
        if !verbosity > 9 then (let srulep = srule_prop srule in let hs = tm_id srulep in Printf.printf "Due to ordering, rejecting srule %s with prop %s\n" hs (s_tm_to_str srulep));
        raise Not_found
    end
  | ShallowPred(al,j,p,b,othl) ->
    let (_,insl) = tm_schema_and_instnames p in
    let maxv = ref Int32.min_int in
    List.iter (fun h -> let v = ordered_salted_v h in if v > !maxv then maxv := v) insl;
    begin
      try
        List.iter (fun q -> ensure_all_instnames_smaller q !maxv) othl
      with Exit ->
        if !verbosity > 9 then (let srulep = srule_prop srule in let hs = tm_id srulep in Printf.printf "Due to ordering, rejecting srule %s with prop %s\n" hs (s_tm_to_str srulep));
        raise Not_found
    end
      
let shallows_computed : (string * bool,unit) Hashtbl.t = Hashtbl.create 100
let shallow_computed : (string,unit) Hashtbl.t = Hashtbl.create 100

let prop_as_shallow_rules a h p b =
  if not (Hashtbl.mem shallows_computed (h,b)) then
    begin
      Hashtbl.add shallows_computed (h,b) ();
      let srules = ref [] in
      let get_func_rule i m1 m2 othl =
	if !use_shallowrules_func && i > 0 then
	  begin
	    if (if !use_shallowrules_func_onlylinear then linear_shallowp i m1 else relevant_shallowp i m1) then
	      begin
		try
		  let (j,m1g,theta) = ms_shallow_linearization m1 in
		  let (thetai,sideeqns) = pinv_subst i theta in
                  if !use_shallowrules_func_onlyalmostlinear && List.length sideeqns > 1 then raise Not_found;
		  let srule = ShallowFunc([a],j,m1g,tmssubst2 m2 thetai 0,sideeqns @ (List.map (fun n -> tmssubst2 n thetai 0) othl)) in
		  if !ordered_shallow_rules then check_ordered_shallow_rule_ok srule;
		  let srulep = srule_prop srule in
		  let hs = tm_id srulep in
		  if not (Hashtbl.mem shallow_computed hs) then
		    begin
		      Hashtbl.add shallow_computed hs ();
		      let (hss,ins) = tm_schema_and_instnames srulep in
                      let hssid = schema_id hss in
		      Hashtbl.add tmid_schema hs (hssid,ins);
                      if !verbosity > 28 then (Printf.printf "%s is id of a shallow rule from %s\nsrule prop: %s\n" hs (tm_id p) (s_tm_to_str srulep); flush stdout);
		      if Hashtbl.mem pos_watch_shallow hs then Printf.printf "(WATCH POS SHALLOW \"%s\" %d)\n" hs !middle_abstract_time;
		      if Hashtbl.mem neg_watch_shallow hs then Printf.printf "(WATCH NEG SHALLOW \"%s\" %d)\n" hs !middle_abstract_time;
		      if Hashtbl.mem pos_rote_shallow hs then
		        srules := (hs,srule)::!srules
		      else if !rote_strict then
			()
		      else if Hashtbl.mem neg_rote_shallow hs then
			()
		      else
			(Hashtbl.add pos_rote_shallow hs (); srules := (hs,srule)::!srules)
		    end
		with Not_found -> ()
	      end
	  end
      in
      let get_pred_rule i p1 othl b = (** if b and p1 or ~b and ~p1 then one of othl must be false **)
	let (h,sp) = tm_head_args p1 in
	match h with
	| TmH(h1) when not (log_const_p h1) || h1 = !eqSet ->
	    if !use_shallowrules_pred && i > 0 then
	      begin
	        if (if !use_shallowrules_pred_onlylinear then linear_shallowp i p1 else relevant_shallowp i p1) then
		  begin
		    try
		      let (j,p1g,theta) = ms_shallow_linearization p1 in
		      let (thetai,sideeqns) = pinv_subst i theta in
                      if !use_shallowrules_pred_onlyalmostlinear && List.length sideeqns > 1 then raise Not_found;
		      let srule = ShallowPred([a],j,p1g,b,sideeqns @ (List.map (fun n -> tmssubst2 n thetai 0) othl)) in
  		      if !ordered_shallow_rules then check_ordered_shallow_rule_ok srule;
		      let srulep = srule_prop srule in
		      let hs = tm_id srulep in
		      if not (Hashtbl.mem shallow_computed hs) then
			begin
			  Hashtbl.add shallow_computed hs ();
			  let (hss,ins) = tm_schema_and_instnames srulep in
                          let hssid = schema_id hss in
			  Hashtbl.add tmid_schema hs (hssid,ins);
			  if !verbosity > 28 then (Printf.printf "%s is id of a shallow rule from %s\nsrule prop: %s\n" hs (tm_id p) (s_tm_to_str srulep); flush stdout);
		          if Hashtbl.mem pos_watch_shallow hs then Printf.printf "(WATCH POS SHALLOW \"%s\" %d)\n" hs !middle_abstract_time;
		          if Hashtbl.mem neg_watch_shallow hs then Printf.printf "(WATCH NEG SHALLOW \"%s\" %d)\n" hs !middle_abstract_time;
			  if Hashtbl.mem pos_rote_shallow hs then
			    srules := (hs,srule)::!srules
			  else if !rote_strict then
			    ()
			  else if Hashtbl.mem neg_rote_shallow hs then
			    ()
			  else
			    (Hashtbl.add pos_rote_shallow hs (); srules := (hs,srule)::!srules)
			end
		    with Not_found -> ()
		  end
	      end
	| _ -> ()
      in
      let rec prop_as_shallow_rules_1 p i othl b =
	if !verbosity > 30 then
	  begin
	    Printf.printf "prop_as_shallow_rules_1\np: %s\ni: %d b: %b\nnothl: %d\n" (s_tm_to_str p) i b (List.length othl);
	    List.iter (fun q -> Printf.printf "Oth: %s\n" (s_tm_to_str q)) othl
	  end;
	match p with
	| All(Set,p1) when b ->
	    if free_in_tm_p p1 0 then
              prop_as_shallow_rules_1 p1 (i+1) (List.map (tmshift 0 1) othl) true
	    else (** skip vacuous quantifier **)
              prop_as_shallow_rules_1 (tmshift 0 (-1) p1) i othl true
	| Ap(TmH(k),Lam(Set,p1)) when k = !exSet && not b ->
	    if free_in_tm_p p1 0 then
              prop_as_shallow_rules_1 p1 (i+1) (List.map (tmshift 0 1) othl) false
	    else (** skip vacuous quantifier **)
              prop_as_shallow_rules_1 (tmshift 0 (-1) p1) i othl false
	| Ap(TpAp(TmH(k),Set),Lam(Set,p1)) when k = !expoly && not b ->
	    if free_in_tm_p p1 0 then
              prop_as_shallow_rules_1 p1 (i+1) (List.map (tmshift 0 1) othl) false
	    else (** skip vacuous quantifier **)
              prop_as_shallow_rules_1 (tmshift 0 (-1) p1) i othl false
	| Ap(Ap(TpAp(TmH(k),Set),m1),m2) when k = !neqPoly ->
	    prop_as_shallow_rules_1 (Ap(Ap(TmH(!eqSet),m1),m2)) i othl (not b)
	| Ap(Ap(TmH(k),m1),m2) when k = !neqSet ->
	    prop_as_shallow_rules_1 (Ap(Ap(TmH(!eqSet),m1),m2)) i othl (not b)
	| Ap(Ap(TpAp(TmH(k),Set),m1),m2) when k = !eqPoly && b ->
	    get_func_rule i m1 m2 othl;
	    get_func_rule i m2 m1 othl;
	    get_pred_rule i p othl (not b)
	| Ap(Ap(TmH(k),m1),m2) when k = !eqSet && b ->
	    get_func_rule i m1 m2 othl;
	    get_func_rule i m2 m1 othl;
	    get_pred_rule i p othl (not b)
	| Imp(p1,TmH(k)) when k = !fal ->
	    prop_as_shallow_rules_1 p1 i othl (not b)
	| Ap(TmH(k),p1) when k = !neg ->
	    prop_as_shallow_rules_1 p1 i othl (not b)
	| Imp(p1,p2) when b ->
	    prop_as_shallow_rules_2 p2 i p1 othl true;
	    prop_as_shallow_rules_2 p1 i (negprop p2) othl false
	| Ap(Ap(TmH(k),p1),p2) when k = !disj && b ->
	    prop_as_shallow_rules_2 p1 i (negprop p2) othl true;
	    prop_as_shallow_rules_2 p2 i (negprop p1) othl true
	| Ap(Ap(TmH(k),p1),p2) when k = !conj && not b ->
	    prop_as_shallow_rules_2 p1 i p2 othl false;
	    prop_as_shallow_rules_2 p2 i p1 othl false
	| Imp(p1,p2) when not b ->
	    prop_as_shallow_rules_1 p2 i othl false;
	    prop_as_shallow_rules_1 p1 i othl true
	| Ap(Ap(TmH(k),p1),p2) when k = !disj && not b ->
	    prop_as_shallow_rules_1 p1 i othl false;
	    prop_as_shallow_rules_1 p2 i othl false
	| Ap(Ap(TmH(k),p1),p2) when k = !conj && b ->
	    prop_as_shallow_rules_1 p1 i othl true;
	    prop_as_shallow_rules_1 p2 i othl true
	| Ap(Ap(TmH(k),p1),p2) when k = !iff ->
	    prop_as_shallow_rules_2 p2 i p1 othl b;
	    prop_as_shallow_rules_2 p1 i p2 othl b;
	    prop_as_shallow_rules_2 p2 i (negprop p1) othl (not b);
	    prop_as_shallow_rules_2 p1 i (negprop p2) othl (not b)
	| _ ->
	    get_pred_rule i p othl (not b)
      and prop_as_shallow_rules_2 p i q othl b =
	prop_as_shallow_rules_1 p i (q::othl) b;
	if !shallow_var_elim then
	  match q with
	  | Ap(Ap(TmH(h),DB(j)),m) when h = !eqSet -> prop_as_shallow_rules_3 p i j m othl b
	  | Ap(Ap(TpAp(TmH(h),Set),DB(j)),m) when h = !eqPoly -> prop_as_shallow_rules_3 p i j m othl b
	  | Ap(Ap(TmH(h),m),DB(j)) when h = !eqSet -> prop_as_shallow_rules_3 p i j m othl b
	  | Ap(Ap(TpAp(TmH(h),Set),m),DB(j)) when h = !eqPoly -> prop_as_shallow_rules_3 p i j m othl b
	  | _ -> ()
      and prop_as_shallow_rules_3 p i j m othl b =
	try
	  let m2 = tmshift j (-1) m in
	  let p2 = tmsubst2 p j m2 0 in
	  let othl2 = List.map (fun q -> tmsubst2 q j m2 0) othl in
	  prop_as_shallow_rules_1 p2 (i-1) othl2 b
	with NegDB -> ()
      in
      prop_as_shallow_rules_1 p 0 [] b;
      !srules
    end
  else
    []
	
let rec tm_delta_norm m =
  match m with
  | Lam(a,m2) -> Lam(a,tm_delta_norm m2)
  | All(a,m2) -> All(a,tm_delta_norm m2)
  | Imp(m1,m2) -> Imp(tm_delta_norm m1,tm_delta_norm m2)
  | Ap(Ap(TpAp(TmH(h),Set),m1),m2) when h = !neqPoly -> Ap(TmH(!neg),tm_delta_norm (Ap(Ap(TmH(!eqSet),m1),m2)))
  | Ap(Ap(TmH(h),m1),m2) when h = !neqSet -> Ap(TmH(!neg),tm_delta_norm (Ap(Ap(TmH(!eqSet),m1),m2)))
  | Ap(Ap(TpAp(TmH(h),a),m1),m2) when h = !neqPoly -> Ap(TmH(!neg),tm_delta_norm (Ap(Ap(TpAp(TmH(!eqSet),a),m1),m2)))
  | _ ->
      let (mh,ms) = tm_head_args m in
      match mh with
      | TmH(h) ->
	  begin
	    try
	      let n = Hashtbl.find funcdef h in
	      tm_delta_norm (tm_app_beta_eta_norm n ms)
	    with Not_found ->
	      try
		let n = Hashtbl.find preddef h in
		tm_delta_norm (tm_app_beta_eta_norm n ms)
	      with Not_found ->
		tm_delta_norm_spine mh ms
	  end
      | _ -> tm_delta_norm_spine mh ms
and tm_delta_norm_spine m s =
  match s with
  | [] -> m
  | n::r -> tm_delta_norm_spine (Ap(m,tm_delta_norm n)) r

exception PfFound

type searchoption =
    ProcessPosProp of string * tm
  | ProcessNegProp of string * tm
  | ProcessInst of string * tm

module PQueue0 = PriorityQueue0 (struct type t = searchoption end)
module PQueue1 = PriorityQueue0 (struct type t = searchoption end)
module PQueue2 = PriorityQueue0 (struct type t = searchoption end)
module PQueue3 = PriorityQueue0 (struct type t = searchoption end)

let atom_count = ref 0

let minisat_modelValue_s i = let v = minisat_modelValue i in if v = 0 then "T" else if v = 1 then "F" else "X"

let add_search_clause c princ =
  if !verbosity > 98 then (Printf.printf "current assignment:\n"; for i = 1 to !atom_count do Printf.printf "%d %s\n" i (minisat_modelValue_s i) done);
  if !verbosity > 6 then (Printf.printf "new search clause:"; List.iter (fun i -> Printf.printf " %d" i) c; Printf.printf "\n");
  clauses := c::!clauses;
  Hashtbl.add clause_princ c princ;
  match c with
  | [] -> raise PfFound
  | [l] ->
    if not (minisat_addClause1 l) then raise PfFound;
    if not (minisat_search()) then raise PfFound
  | [l1;l2] ->
    if not (minisat_addClause2 l1 l2) then raise PfFound;
    if not (minisat_search()) then raise PfFound
  | [l1;l2;l3] ->
    if not (minisat_addClause3 l1 l2 l3) then raise PfFound;
    if not (minisat_search()) then raise PfFound
  | _ ->
    List.iter (fun l -> minisat_addLit l) c;
    if not (minisat_addClause()) then raise PfFound;
    if not (minisat_search()) then raise PfFound

let atomh : (string,int) Hashtbl.t = Hashtbl.create 100
let atomh_rev : (int,string) Hashtbl.t = Hashtbl.create 100
let atomh_prop_rev : (int,tm) Hashtbl.t = Hashtbl.create 100
let atomh_sch_rev : (int,string) Hashtbl.t = Hashtbl.create 100
let posproph : (string,unit) Hashtbl.t = Hashtbl.create 100
let negproph : (string,unit) Hashtbl.t = Hashtbl.create 100
let insth : (string,unit) Hashtbl.t = Hashtbl.create 100
let pospropdoneh : (string,unit) Hashtbl.t = Hashtbl.create 100
let negpropdoneh : (string,unit) Hashtbl.t = Hashtbl.create 100
let instdoneh : (string,unit) Hashtbl.t = Hashtbl.create 100

let instlist : (tm * string) list ref = ref []
let atmshlinh : (int * tm * bool,int * tm list * string list * string list) Hashtbl.t = Hashtbl.create 100
let trmshlinh : (int * tm,int * tm list * tm * string) Hashtbl.t = Hashtbl.create 100
let posatmh : (tm,int * (tm list) * string) Hashtbl.t = Hashtbl.create 100
let negatmh : (tm,int * (tm list) * string) Hashtbl.t = Hashtbl.create 100
let poseqnlist : (int * tm * tm * string) list ref = ref []
let negeqnlist : (int * tm * tm * string) list ref = ref []
let alllist : (int * bool * tm * string) list ref = ref []

let new_atomh h =
  try
    Hashtbl.find atomh h
  with Not_found ->
    incr atom_count;
    Hashtbl.add atomh h !atom_count;
    Hashtbl.add atomh_rev !atom_count h;
    if !verbosity > 6 then Printf.printf "atom %d %s\n" !atom_count h;
    !atom_count

let new_atom p =
  let h = tm_id p in
  try
    Hashtbl.find atomh h
  with Not_found ->
    incr atom_count;
    Hashtbl.add atomh h !atom_count;
    Hashtbl.add atomh_rev !atom_count h;
    Hashtbl.add atomh_prop_rev !atom_count p;
    if !verbosity > 6 then Printf.printf "atom %d %s %s\n" !atom_count h (s_tm_to_str p);
    !atom_count

let rec new_literal p =
  match p with
  | Imp(p1,All(Prop,DB(0))) -> - new_literal p1
  | Imp(p1,TmH(h)) when h = !fal -> - new_literal p1
  | Ap(TmH(h),p1) when h = !neg -> - new_literal p1
  | Ap(Ap(TmH(h),m1),m2) when h = !neqSet -> - new_literal (Ap(Ap(TmH(!eqSet),m1),m2))
  | Ap(Ap(TpAp(TmH(h),Set),m1),m2) when h = !neqSet -> - new_literal (Ap(Ap(TmH(!eqSet),m1),m2))
  | _ -> new_atom p

let rec tm_size m =
  match m with
  | Ap(m1,m2) -> 1 + tm_size m1 + tm_size m2
  | Lam(a,m1) -> 1 + tm_size m1
  | Imp(m1,m2) -> 1 + tm_size m1 + tm_size m2
  | All(a,m1) -> 1 + tm_size m1
  | _ -> 1

let rec tm_dist m n d =
  if m = n then
    0
  else if d > 0 then
    let (h1,s1) = tm_head_args m in
    let (h2,s2) = tm_head_args n in
    if h1 = h2 then
      spine_dist s1 s2 (d-1)
    else
      1 lsl (2*d)
  else
    1
and spine_dist s1 s2 d =
  match (s1,s2) with
  | ([],[]) -> 0
  | (m::sr1,n::sr2) -> max (tm_dist m n d) (spine_dist sr1 sr2 d)
  | _ -> 128

let rec truthiness p =
  match p with
  | Ap(Ap(TmH(k),m),n) -> tm_dist m n 4 (** the more the terms look alike, the more truthy the equation is **)
  | Ap(Ap(TpAp(TmH(k),_),m),n) when k = !eqPoly -> tm_dist m n 4
  | Imp(p1,TmH(h)) when h = !fal -> falsiness p1
  | Ap(TmH(h),p1) when h = !neg -> falsiness p1
  | TmH(h) when h = !tru -> 0
  | _ -> 100
and falsiness p =
  match p with
  | TmH(h) when h = !fal -> 0
  | Imp(p1,TmH(h)) when h = !fal -> truthiness p1
  | Ap(TmH(h),p1) when h = !neg -> truthiness p1
  | _ -> 100

let pos_prop_priority_1 h m =
  let pr = falsiness m in
  match !prioritysalt with
  | Some(s) -> min pr (Hashaux.hexsubstring_int8 (Hash.hashval_hexstring (Hash.hashtag (Hash.hexstring_hashval h) s)) 0)
  | None -> pr

let pos_prop_priority_2 h m =
  tm_size m

let neg_prop_priority_1 h m =
  let pr = truthiness m in
  match !prioritysalt with
  | Some(s) -> min pr (Hashaux.hexsubstring_int8 (Hash.hashval_hexstring (Hash.hashtag (Hash.hexstring_hashval h) s)) 0)
  | None -> pr

let neg_prop_priority_2 h m =
  tm_size m

let inst_priority_1 h m =
  match !prioritysalt with
  | Some(s) -> Hashaux.hexsubstring_int8 (Hash.hashval_hexstring (Hash.hashtag (Hash.hexstring_hashval h) s)) 0
  | None -> 100

let inst_priority_2 h m =
  tm_size m

let propositional_p p =
  match p with
  | Imp(_,_) -> true
  | Ap(TmH(h),_) when h = !neg -> true
  | Ap(Ap(TmH(k),_),_) when k = !conj || k = !disj || k = !iff -> true
  | _ -> false

let possibly_new_pos_prop p shl phl nhl ihl =
  if !middle_search || (!prop_endphase && propositional_p p) then
    let h = tm_id p in
    if not (Hashtbl.mem posproph h) then
      begin
        Hashtbl.add posproph h ();
        let (ps,ins) = tm_schema_and_instnames p in
        let hs = schema_id ps in
        Hashtbl.add tmid_schema h (hs,ins);
        if !verbosity > 19 then Printf.printf "New pos %s [%s] %s\n" h hs (s_tm_to_str p);
        List.iter (fun sh -> Hashtbl.add pos_src_shallow h sh) shl;
        List.iter (fun ph -> Hashtbl.add pos_src_pos h ph) phl;
        List.iter (fun nh -> Hashtbl.add pos_src_neg h nh) nhl;
        List.iter (fun ih -> Hashtbl.add pos_src_inst h ih) ihl;
        if Hashtbl.mem pos_watch_pos h then Printf.printf "(WATCH POS POS \"%s\" %d)\n" h !middle_abstract_time;
        if Hashtbl.mem neg_watch_pos h then Printf.printf "(WATCH NEG POS \"%s\" %d)\n" h !middle_abstract_time;
        if Hashtbl.mem pos_rote_pos h || (!using_guidance && pos_guidance_prop_2 true p ps ins) then
          begin
            if !verbosity > 19 then Printf.printf "priority 0 0 due to pos rote pos\n";
            PQueue0.insertWithPriority 0 (ProcessPosProp(h,p))
          end
        else if !rote_strict then
          ()
        else if Hashtbl.mem neg_rote_pos h || (!using_guidance && neg_guidance_prop_2 true p ps ins) then
          begin
            if !verbosity > 19 then Printf.printf "priority 2 1000 due to neg rote pos\n";
            PQueue2.insertWithPriority 1000 (ProcessPosProp(h,p))
          end
        else
	  let pr1 = pos_prop_priority_1 h p in
	  let pr2 = pos_prop_priority_2 h p in
	  begin
	    if !verbosity > 19 then Printf.printf "pos prop %s priority %d in queue 1 and %d in queue 2\n" h pr1 pr2;
	    PQueue1.insertWithPriority pr1 (ProcessPosProp(h,p));
	    PQueue2.insertWithPriority pr2 (ProcessPosProp(h,p))
	  end
      end

let possibly_new_neg_prop p shl phl nhl ihl =
  if !middle_search || (!prop_endphase && propositional_p p) then
    let h = tm_id p in
    if not (Hashtbl.mem negproph h) then
      begin
        Hashtbl.add negproph h ();
        let (ps,ins) = tm_schema_and_instnames p in
        let hs = schema_id ps in
        Hashtbl.add tmid_schema h (hs,ins);
        if !verbosity > 19 then Printf.printf "New neg %s [%s] %s\n" h hs (s_tm_to_str p);
        List.iter (fun sh -> Hashtbl.add neg_src_shallow h sh) shl;
        List.iter (fun ph -> Hashtbl.add neg_src_pos h ph) phl;
        List.iter (fun nh -> Hashtbl.add neg_src_neg h nh) nhl;
        List.iter (fun ih -> Hashtbl.add neg_src_inst h ih) ihl;
        if Hashtbl.mem pos_watch_neg h then Printf.printf "(WATCH POS NEG \"%s\" %d)\n" h !middle_abstract_time;
        if Hashtbl.mem neg_watch_neg h then Printf.printf "(WATCH NEG NEG \"%s\" %d)\n" h !middle_abstract_time;
        if Hashtbl.mem pos_rote_neg h || (!using_guidance && pos_guidance_prop_2 false p ps ins) then
          begin
            if !verbosity > 19 then Printf.printf "priority 0 0 due to pos rote neg\n";
            PQueue0.insertWithPriority 0 (ProcessNegProp(h,p))
          end
        else if !rote_strict then
          ()
        else if Hashtbl.mem neg_rote_neg h || (!using_guidance && neg_guidance_prop_2 false p ps ins) then
          begin
            if !verbosity > 19 then Printf.printf "priority 2 1000 due to neg rote neg\n";
            PQueue2.insertWithPriority 1000 (ProcessNegProp(h,p))
          end
        else
	  begin
	    match p with
	    | Ap(Ap(TmH(k),m),n) when k = !eqSet && m = n ->
	       PQueue0.insertWithPriority 0 (ProcessNegProp(h,p))
	    | Ap(Ap(TpAp(TmH(k),Set),m),n) when k = !eqPoly && m = n ->
	       PQueue0.insertWithPriority 0 (ProcessNegProp(h,p))
	    | _ ->
	       let pr1 = neg_prop_priority_1 h p in
	       let pr2 = neg_prop_priority_2 h p in
	       if !verbosity > 19 then Printf.printf "neg prop %s priority %d in queue 1 and %d in queue 2\n" h pr1 pr2;
	       PQueue1.insertWithPriority pr1 (ProcessNegProp(h,p));
	       PQueue2.insertWithPriority pr2 (ProcessNegProp(h,p))
	  end
      end
        
let possibly_new_neg_prop_priority p pr =
  let h = tm_id p in
  if not (Hashtbl.mem negproph h) then
    begin
      Hashtbl.add negproph h ();
      if !verbosity > 19 then Printf.printf "New neg %s %s priority %d\n" h (s_tm_to_str p) pr;
      if pr < 50 then
	PQueue0.insertWithPriority pr (ProcessNegProp(h,p))
      else if pr < 500 then
	PQueue1.insertWithPriority pr (ProcessNegProp(h,p))
      else
	PQueue2.insertWithPriority pr (ProcessNegProp(h,p))
    end
      
let possibly_new_pos_prop_priority p pr =
  let h = tm_id p in
  if not (Hashtbl.mem posproph h) then
    begin
      Hashtbl.add posproph h ();
      if !verbosity > 19 then Printf.printf "New pos %s %s priority %d\n" h (s_tm_to_str p) pr;
      if pr < 50 then
	PQueue0.insertWithPriority pr (ProcessPosProp(h,p))
      else if pr < 500 then
	PQueue1.insertWithPriority pr (ProcessPosProp(h,p))
      else
	PQueue2.insertWithPriority pr (ProcessPosProp(h,p))
    end

let rec possibly_new_prop p b shl phl nhl ihl =
  match p with
  | Imp(p1,All(Prop,DB(0))) -> possibly_new_prop p1 (not b) shl phl nhl ihl
  | Imp(p1,TmH(h)) when h = !fal -> possibly_new_prop p1 (not b) shl phl nhl ihl
  | Ap(TmH(h),p1) when h = !neg -> possibly_new_prop p1 (not b) shl phl nhl ihl
  | Ap(Ap(TpAp(TmH(h),Set),m1),m2) when h = !neqPoly -> possibly_new_prop (Ap(Ap(TmH(!eqSet),m1),m2)) (not b) shl phl nhl ihl
  | Ap(Ap(TmH(h),m1),m2) when h = !neqSet -> possibly_new_prop (Ap(Ap(TmH(!eqSet),m1),m2)) (not b) shl phl nhl ihl
  | Ap(Ap(TpAp(TmH(h),a),m1),m2) when h = !neqPoly -> possibly_new_prop (Ap(Ap(TpAp(TmH(!eqPoly),a),m1),m2)) (not b) shl phl nhl ihl
  | _ -> if b then possibly_new_pos_prop p shl phl nhl ihl else possibly_new_neg_prop p shl phl nhl ihl

let possibly_new_inst m nh =
  if !middle_search then
    let h = tm_id m in
    if not (Hashtbl.mem insth h) then
      begin
        Hashtbl.add insth h ();
        let (ms,ins) = tm_schema_and_instnames m in
        let hs = schema_id ms in
        Hashtbl.add tmid_schema h (hs,ins);
        if !verbosity > 19 then Printf.printf "New inst %s [%s] %s\n" h hs (s_tm_to_str m);
        Hashtbl.add inst_src_neg h nh;
        if Hashtbl.mem pos_watch_pos h then Printf.printf "(WATCH POS INST \"%s\" %d)\n" h !middle_abstract_time;
        if Hashtbl.mem neg_watch_pos h then Printf.printf "(WATCH NEG INST \"%s\" %d)\n" h !middle_abstract_time;
        if Hashtbl.mem pos_rote_inst h || (!using_guidance && pos_guidance_inst_3 m ms ins hs) then
          begin
            if !verbosity > 19 then Printf.printf "priority 0 0 due to pos rote inst\n";
            PQueue0.insertWithPriority 0 (ProcessInst(h,m))
          end
        else if !rote_strict then
          ()
        else if Hashtbl.mem neg_rote_inst h || (!using_guidance && neg_guidance_inst_3 m ms ins hs) then
          begin
            if !verbosity > 19 then Printf.printf "priority 2 1000 due to pos rote inst\n";
            PQueue2.insertWithPriority 1000 (ProcessInst(h,m))
          end
        else
	  begin
	    PQueue1.insertWithPriority (inst_priority_1 h m) (ProcessInst(h,m));
	    PQueue2.insertWithPriority (inst_priority_2 h m) (ProcessInst(h,m))
	  end
      end
        
let shallowclause : (int list,string) Hashtbl.t = Hashtbl.create 100

let rec split_spine s pos =
  match s with
  | m::sr ->
      if pos > 0 then
	let (x,y,z) = split_spine sr (pos-1) in
	(m::x,y,z)
      else
	([],m,sr)
  | [] -> raise Not_found
  
let register_shallow_rule (hss,srule) =
  match srule with
  | ShallowFunc(al,j,m,m3,othl) ->
      let (posl,negl) = neg_split_by_polarity othl in
      let al2 = List.map (fun i -> -i) al in
      Hashtbl.add lin_shallow_func_rules (j,m) (hss,al2,m3,posl,negl);
      List.iter
	(fun (i,theta,m2,h) ->
	  let e2 = Ap(Ap(TmH(!eqSet),tmssubst m3 theta 0),m2) in
	  let ib = -(new_atom e2) in
	  possibly_new_prop e2 false [hss] [] [h] [];
	  let ll = ref (ib::al) in
	  List.iter
	    (fun q ->
	      let q = tmssubst q theta 0 in
	      let k = new_atom q in
	      possibly_new_prop q true [hss] [] [h] [];
	      ll := k::!ll)
	    posl;
	  List.iter
	    (fun q ->
	      let q = tmssubst q theta 0 in
	      let k = -(new_atom q) in
	      possibly_new_prop q false [hss] [] [h] [];
	      ll := k::!ll)
	    negl;
	  Hashtbl.add shallowclause (i::!ll) hss;
	  add_search_clause (i::!ll) [i])
	(Hashtbl.find_all trmshlinh (j,m))
  | ShallowPred(al,j,p,b,othl) ->
      let (posl,negl) = neg_split_by_polarity othl in
      let al2 = List.map (fun i -> -i) al in
      Hashtbl.add lin_shallow_pred_rules (j,p,b) (hss,al2,posl,negl);
      List.iter
	(fun (i,theta,ph,nh) ->
	  let ll = ref al2 in
	  List.iter
	    (fun q ->
	      let q = tmssubst q theta 0 in
	      let k = new_atom q in
	      possibly_new_prop q true [hss] ph nh [];
	      ll := k::!ll)
	    posl;
	  List.iter
	    (fun q ->
	      let q = tmssubst q theta 0 in
	      let k = -(new_atom q) in
	      possibly_new_prop q false [hss] ph nh [];
	      ll := k::!ll)
	    negl;
	  Hashtbl.add shallowclause (i::!ll) hss;
	  add_search_clause (i::!ll) [i])
	(Hashtbl.find_all atmshlinh (j,p,b))

let reset_shallow_rules () =
  Hashtbl.clear shallowclause;
  Hashtbl.clear lin_shallow_func_rules;
  Hashtbl.clear lin_shallow_pred_rules;
  Hashtbl.clear shallow_func_rules;
  Hashtbl.clear shallow_func_func_rules;
  Hashtbl.clear shallow_pos_pred_rules;
  Hashtbl.clear shallow_neg_pred_rules;
  Hashtbl.clear shallow_pos_pred_func_rules;
  Hashtbl.clear shallow_neg_pred_func_rules;
  Hashtbl.clear shallows_computed;
  Hashtbl.clear shallow_computed

let noinsts = ref true

let possibly_new_inst_init m =
  let h = tm_id m in
  if not (Hashtbl.mem insth h) then
    begin
      noinsts := false;
      Hashtbl.add insth h ();
      let (ms,ins) = tm_schema_and_instnames m in
      let hs = schema_id ms in
      Hashtbl.add tmid_schema h (hs,ins);
      if !verbosity > 19 then Printf.printf "New inst %s [%s] %s\n" h hs (s_tm_to_str m);
      if Hashtbl.mem pos_watch_pos h then Printf.printf "(WATCH POS INST \"%s\" %d)\n" h !middle_abstract_time;
      if Hashtbl.mem neg_watch_pos h then Printf.printf "(WATCH NEG INST \"%s\" %d)\n" h !middle_abstract_time;
      if Hashtbl.mem pos_rote_inst h || (!using_guidance && pos_guidance_inst_3 m ms ins hs) then
        begin
          if !verbosity > 19 then Printf.printf "priority 0 0 due to pos rote inst\n";
          PQueue0.insertWithPriority 0 (ProcessInst(h,m))
        end
      else if !rote_strict then
        ()
      else if Hashtbl.mem neg_rote_inst h || (!using_guidance && neg_guidance_inst_3 m ms ins hs) then
        begin
          if !verbosity > 19 then Printf.printf "priority 2 1000 due to pos rote inst\n";
          PQueue2.insertWithPriority 1000 (ProcessInst(h,m))
        end
      else
	begin
	  PQueue1.insertWithPriority (inst_priority_1 h m) (ProcessInst(h,m));
	  PQueue2.insertWithPriority (inst_priority_2 h m) (ProcessInst(h,m))
	end
    end

let possibly_new_inst_priority m pr =
  let h = tm_id m in
  if not (Hashtbl.mem insth h) then
    begin
      Hashtbl.add insth h ();
      if !verbosity > 19 then Printf.printf "New inst %s %s priority %d\n" h (s_tm_to_str m) pr;
      if pr < 50 then
	PQueue0.insertWithPriority pr (ProcessInst(h,m))
      else if pr < 500 then
	PQueue1.insertWithPriority pr (ProcessInst(h,m))
      else
	PQueue2.insertWithPriority pr (ProcessInst(h,m))
    end

    (*** Following Nelson Oppen 1978 Congruence Closure Paper ***)
let enablecongclos = ref true
let congdebug = ref false
let congeqnsh : (int,unit) Hashtbl.t = Hashtbl.create 10
let congeqnlits = ref []
let congeqnhs = ref []
let congdeqnsh : (int,unit) Hashtbl.t = Hashtbl.create 10
type cctables = (tm,unit) Hashtbl.t * (tm,tm) Hashtbl.t * (tm,tm) Hashtbl.t * (tm,tm) Hashtbl.t
let cc : cctables = (Hashtbl.create 10,Hashtbl.create 10,Hashtbl.create 10,Hashtbl.create 10)

let print_cctables cc =
  let (ccv,ccp,ccb,ccf) = cc in
  Printf.printf "===== CC ======\n";
  Printf.printf "===== handled ======\n";
  Hashtbl.iter (fun k () -> Printf.printf "%s\n" (s_tm_to_str k)) ccv;
  Printf.printf "===== pred ======\n";
  Hashtbl.iter (fun k v -> Printf.printf "%s  ~~>  %s\n" (s_tm_to_str k) (s_tm_to_str v)) ccp;
  Printf.printf "===== back ======\n";
  Hashtbl.iter (fun k v -> Printf.printf "%s  ~~>  %s\n" (s_tm_to_str k) (s_tm_to_str v)) ccb;
  Printf.printf "===== find ======\n";
  Hashtbl.iter (fun k v -> Printf.printf "%s  ~~>  %s\n" (s_tm_to_str k) (s_tm_to_str v)) ccf

let rec cong_find ccf m =
  try
    let r = cong_find ccf (Hashtbl.find ccf m) in
    Hashtbl.replace ccf m r;
    r
  with Not_found ->
    m

let cong_union (ccb,ccf) m n =
  let m1 = cong_find ccf m in
  let n1 = cong_find ccf n in
  Hashtbl.add ccf m1 n1;
  Hashtbl.add ccb n1 m1

let cong_class ccb m =
  let mh : (tm,unit) Hashtbl.t = Hashtbl.create 10 in
  let rec cong_class_1 m =
    if not (Hashtbl.mem mh m) then
      begin
	Hashtbl.add mh m ();
	List.iter (fun n -> cong_class_1 n) (Hashtbl.find_all ccb m)
      end
  in
  cong_class_1 m;
  let ml = ref [] in
  Hashtbl.iter (fun m () -> ml := m::!ml) mh;
  !ml

let cong_preds (ccp,ccb) m =
  let mh : (tm,unit) Hashtbl.t = Hashtbl.create 10 in
  let ph : (tm,unit) Hashtbl.t = Hashtbl.create 10 in
  let rec cong_preds_1 m =
    if not (Hashtbl.mem mh m) then
      begin
	List.iter (fun p -> if not (Hashtbl.mem ph p) then Hashtbl.add ph p ()) (Hashtbl.find_all ccp m);
	Hashtbl.add mh m ();
	List.iter (fun n -> cong_preds_1 n) (Hashtbl.find_all ccb m)
      end
  in
  cong_preds_1 m;
  let pl = ref [] in
  Hashtbl.iter (fun p () -> pl := p::!pl) ph;
  !pl

let rec cong_merge (ccp,ccb,ccf) m n =
  if !congdebug then Printf.printf "merging %s with %s\n" (s_tm_to_str m) (s_tm_to_str n);
  let m1 = cong_find ccf m in
  let n1 = cong_find ccf n in
  if m1 = n1 then
    ()
  else
    let pml = cong_preds (ccp,ccb) m1 in
    let pnl = cong_preds (ccp,ccb) n1 in
    cong_union (ccb,ccf) m1 n1;
    List.iter
      (fun p ->
	let p1 = cong_find ccf p in
	(List.iter
	   (fun q ->
	     let q1 = cong_find ccf q in
	     if not (p1 = q1) && cong_congruent (ccp,ccb,ccf) p q then
	       cong_merge (ccp,ccb,ccf) p q)
	   pnl))
      pml
and cong_congruent (ccp,ccb,ccf) m n =
  let (mh,ms) = tm_head_args m in
  let (nh,ns) = tm_head_args n in
  if mh = nh then
    begin
      try
	if not (List.length ms = List.length ns) then
	  begin
	    Printf.printf "#+#++ cong_congruent iter2 |ms| = %d ; |ns| = %d ;\nm = %s\nn = %s\n" (List.length ms) (List.length ns) (s_tm_to_str m) (s_tm_to_str n);
	    raise Exit
	  end
	else
	  List.iter2
	    (fun mi ni ->
	      let mi1 = cong_find ccf mi in
	      let ni1 = cong_find ccf ni in
	      if not (mi1 = ni1) then raise Exit)
	    ms ns;
	true
      with Exit -> false
    end
  else
    false

let rec extend_cong_graph (ccv,ccp,ccb,ccf) m =
  if not (Hashtbl.mem ccv m) then
    begin
      if !congdebug then Printf.printf "extend_cong_graph %s\n" (s_tm_to_str m);
      let (mh,s) = tm_head_args m in
      List.iter
	(fun n ->
	  extend_cong_graph (ccv,ccp,ccb,ccf) n;
	  if !congdebug then Printf.printf "back from handling arg %s to extend_cong_graph %s\n" (s_tm_to_str n) (s_tm_to_str m);
	  let pnl = cong_preds (ccp,ccb) (cong_find ccf n) in
	  List.iter
	    (fun p ->
	      if !congdebug then Printf.printf "pred %s in extend_cong_graph %s\n" (s_tm_to_str p) (s_tm_to_str m);
	      let p1 = cong_find ccf p in
	      if !congdebug then Printf.printf "pred find %s\n" (s_tm_to_str p1);
	      let m1 = cong_find ccf m in
	      if !congdebug then Printf.printf "m find %s\n" (s_tm_to_str m1);
	      if not (p1 = m1) && cong_congruent (ccp,ccb,ccf) p m then
		cong_merge (ccp,ccb,ccf) p m)
	    pnl;
	  Hashtbl.add ccp n m)
	s;
      Hashtbl.add ccv m ()
    end

let cong m1 m2 =
  extend_cong_graph cc m1;
  extend_cong_graph cc m2;
  let (_,_,_,ccf) = cc in
  let n1 = cong_find ccf m1 in
  let n2 = cong_find ccf m2 in
  if !congdebug then
    begin
      if not (m1 = n1) then Printf.printf "cong_find\n   %s\n ~> %s\n" (s_tm_to_str m1) (s_tm_to_str n1);
      if not (m2 = n2) then Printf.printf "cong_find\n   %s\n ~> %s\n" (s_tm_to_str m2) (s_tm_to_str n2);
    end;
  n1 = n2

    (* h is id of p, i is -literal for p, b is true iff p is positive *)
let process_pred_lin_shallow h i p b =
  let ph = if b then [h] else [] in
  let nh = if b then [] else [h] in
  let shl = shallow_linearizations p in
  List.iter
    (fun (j,q,theta) ->
      Hashtbl.add atmshlinh (j,q,b) (i,theta,ph,nh);
      List.iter
	(fun (hss,al,posl,negl) ->
	  if !verbosity > 19 then Printf.printf "Applying pred lin %s with %b %s\nj %d q %s\n" hss b (s_tm_to_str p) j (s_tm_to_str q);
	  let ll = ref al in
	  List.iter
	    (fun q ->
	      let q = tmssubst q theta 0 in
	      let k = new_atom q in
	      possibly_new_prop q true [hss] ph nh [];
	      ll := k::!ll)
	    posl;
	  List.iter
	    (fun q ->
	      let q = tmssubst q theta 0 in
	      let k = -(new_atom q) in
	      possibly_new_prop q false [hss] ph nh [];
	      ll := k::!ll)
	    negl;
	  Hashtbl.add shallowclause (i::!ll) hss;
	  add_search_clause (i::!ll) [i])
	(Hashtbl.find_all lin_shallow_pred_rules (j,q,b)))
    shl

  (* h is id of m1 = m2 or m2 = m1, i is positive literal for the eqn *)
  (* b is false if we're applying to a neg eqn (the default)
     b is true if we're applying to a pos eqn (optional if apply_funcrules_poseqns true) *)
let process_func_lin_shallow h i m1 m2 b =
  let i = if b then -i else i in
  List.iter
    (fun (j,m1g,theta) ->
      Hashtbl.add trmshlinh (j,m1g) (i,theta,m2,h);
      List.iter
	(fun (hss,al,m3,posl,negl) ->
	  if !verbosity > 19 then Printf.printf "Applying func lin %s\n" hss;
	  let e2 = Ap(Ap(TmH(!eqSet),tmssubst m3 theta 0),m2) in
	  let ib = if b then new_atom e2 else -(new_atom e2) in
	  possibly_new_prop e2 b [hss] [] [h] [];
	  let ll = ref (ib::al) in
	  List.iter
	    (fun q ->
	      let q = tmssubst q theta 0 in
	      let k = new_atom q in
	      possibly_new_prop q true [hss] [] [h] [];
	      ll := k::!ll)
	    posl;
	  List.iter
	    (fun q ->
	      let q = tmssubst q theta 0 in
	      let k = -(new_atom q) in
	      possibly_new_prop q false [hss] [] [h] [];
	      ll := k::!ll)
	    negl;
	  Hashtbl.add shallowclause (i::!ll) hss;
	  add_search_clause (i::!ll) [i])
	(Hashtbl.find_all lin_shallow_func_rules (j,m1g)))
    (shallow_linearizations m1)

let rec process_pos_prop q pr h p =
  if !verbosity > 29 then (Printf.printf "Maybe Processing pos %s %s\n" h (s_tm_to_str p); flush stdout);
  if q = 0 || q = 3 then (** don't use the prop model to delay very high priority or very low priority options **)
    process_pos_prop_2 h p
  else
    let i = new_atomh h in
    let v = minisat_modelValue i in
    if !verbosity > 29 then (Printf.printf "modelValue %d\n" v; flush stdout);
    if v > 0 then (** delay working on it if it's not on the branch identified by the current minisat assignment **)
      if q = 1 then
        if pr < 256 then
          PQueue1.insertWithPriority (pr*2+1) (ProcessPosProp(h,p))
        else
          PQueue2.insertWithPriority pr (ProcessPosProp(h,p))
      else if pr < 2048 then
        PQueue2.insertWithPriority (pr*2+1) (ProcessPosProp(h,p))
      else
        PQueue3.insertWithPriority pr (ProcessPosProp(h,p))
    else
      process_pos_prop_2 h p
and process_pos_prop_2 h p =
  if !verbosity > 4 then Printf.printf "Processing pos %s %s\n" h (s_tm_to_str p);
  if !verbosity > 9 then Printf.printf "(PROCESSPOS \"%s\" %d %d)\n" h !abstract_time !middle_abstract_time;
  incr_abstract_time();
  process_pos_prop_3 h p
and process_pos_prop_3 h p =
  Hashtbl.add pospropdoneh h ();
  if !using_guidance then
    begin
      let (sch,il) = tm_schema_and_instnames p in
      let schid = schema_id sch in
      let ct = ProcessedPos(schid) in
      Hashtbl.add condstable (ct,il) ();
      apply_forward_guidance_rules ct il;
      List.iter
        (fun q ->
          possibly_new_pos_prop q [] [h] [] [];
          possibly_new_neg_prop q [] [h] [] [])
        !guidance_cuts;
      List.iter
        (fun m -> possibly_new_inst m h)
        !guidance_insttrms;
      guidance_cuts := [];
      guidance_insttrms := [];
    end;
  match p with
  | Imp(p1,p2) ->
      possibly_new_prop p1 false [] [h] [] [];
      possibly_new_prop p2 true [] [h] [] [];
      add_search_clause [-(new_atomh h); -(new_literal p1); new_literal p2] [-(new_atomh h)]
  | All(Set,p1) ->
    let i = - new_atomh h in
    List.iter
      (fun (m,ih) ->
        let p1m = tmsubst p1 0 m in
        Hashtbl.add clause_inst [i; new_literal p1m] m;
        possibly_new_prop p1m true [] [h] [] [ih];
        add_search_clause [i; new_literal p1m] [i])
      !instlist;
    if !dynamic_shallowrules then
      begin
	let srules = prop_as_shallow_rules (-i) h p true in
	if not (srules = []) then
	  begin
	    if !verbosity > 25 then (Printf.printf "dynamic srules %d %s\n" (List.length srules) h; flush stdout);
	    Hashtbl.add shallowatom (-i) ();
	    List.iter
	      (fun (hss,srule) ->
		Hashtbl.add shallow_src_pos hss h;
		register_shallow_rule (hss,srule))
	      srules
	  end
      end;
    alllist := (i,true,p1,h) :: !alllist
  | Ap(Ap(TpAp(TmH(k),Set),n1),n2) when k = !eqPoly -> process_pos_prop_3 h (Ap(Ap(TmH(!eqSet),n1),n2))
  | Ap(Ap(TmH(k),n1),n2) when k = !eqSet && n1 = n2 -> () (** ignore trivial eqns **)
  | Ap(Ap(TmH(k),n1),n2) when k = !eqSet ->
      let j = -(new_atomh h) in
      if !enablecongclos then
	begin
	  let msvj = minisat_search1 j in
	  if !verbosity > 19 then Printf.printf "trying cong clos = for eqn %d %b\n" j msvj;
	  if not msvj then (** the eqn is on every open path **)
	    begin
	      Hashtbl.add congeqnsh j ();
	      congeqnlits := j::!congeqnlits;
	      congeqnhs := h::!congeqnhs;
	      let (ccv,ccp,ccb,ccf) = cc in
	      extend_cong_graph cc n1;
	      extend_cong_graph cc n2;
	      cong_merge (ccp,ccb,ccf) n1 n2;
	      (** now since a new eqn has been added check if any more diseqns can be refuted **)
	      List.iter
		(fun (i,m1,m2,neqh) ->
		  if not (Hashtbl.mem congdeqnsh i) && minisat_modelValue (-i) = 0 then (** deqn on current open branch **)
		    begin
		      if !verbosity > 19 then Printf.printf "testing cong clos for %d: %s =? %s\n" i (s_tm_to_str m1) (s_tm_to_str m2);
		      if cong m1 m2 then
			begin
			  if !verbosity > 19 then Printf.printf "cong clos gave clause for %d\n" i;
			  Hashtbl.add congdeqnsh i ();
			  add_search_clause (i::!congeqnlits) (i::!congeqnlits) (** consider them all to be principal **)
			end
		    end)
		!negeqnlist
	    end
	end;
      if !eqn_sym_rule then
        begin
          let q = Ap(Ap(TmH(!eqSet),n2),n1) in
          let qh = tm_id q in
          let i = new_atomh qh in
          possibly_new_prop q true [] [h] [] [];
          add_search_clause [j;i] [j];
        end;
      List.iter
	(fun (i,m1,m2,neqh) ->
          if (not (i = -j) || !self_confrontations) then
            begin
              let d1 = Ap(Ap(TmH(!eqSet),m1),n1) in
              let d2 = Ap(Ap(TmH(!eqSet),m1),n2) in
              let d3 = Ap(Ap(TmH(!eqSet),m2),n1) in
              let d4 = Ap(Ap(TmH(!eqSet),m2),n2) in
              possibly_new_prop d1 false [] [h] [neqh] [];
              possibly_new_prop d2 false [] [h] [neqh] [];
              possibly_new_prop d3 false [] [h] [neqh] [];
              possibly_new_prop d4 false [] [h] [neqh] [];
              add_search_clause [i; j; -(new_atom d1); -(new_atom d3)] [i;j];
              add_search_clause [i; j; -(new_atom d1); -(new_atom d4)] [i;j];
              add_search_clause [i; j; -(new_atom d2); -(new_atom d3)] [i;j];
              add_search_clause [i; j; -(new_atom d2); -(new_atom d4)] [i;j]
            end)
	!negeqnlist;
      process_pred_lin_shallow h j p true;
      if !apply_funcrules_poseqns then
        begin
	  process_func_lin_shallow h (-j) n1 n2 true;
          process_func_lin_shallow h (-j) n2 n1 true
	end;
      poseqnlist := (j,n1,n2,h)::!poseqnlist
  | TmH(k) when k = !fal -> add_search_clause [-(new_atomh h)] [-(new_atomh h)]
  | TmH(k) when k = !tru -> ()
  | Ap(Ap(TmH(k),p1),p2) when k = !conj ->
      possibly_new_prop p1 true [] [h] [] [];
      possibly_new_prop p2 true [] [h] [] [];
      add_search_clause [-(new_atomh h); new_literal p1] [-(new_atomh h)];
      add_search_clause [-(new_atomh h); new_literal p2] [-(new_atomh h)]
  | Ap(Ap(TmH(k),p1),p2) when k = !disj ->
      possibly_new_prop p1 true [] [h] [] [];
      possibly_new_prop p2 true [] [h] [] [];
      add_search_clause [-(new_atomh h); new_literal p1; new_literal p2] [-(new_atomh h)]
  | Ap(Ap(TmH(k),p1),p2) when k = !iff ->
      possibly_new_prop p1 false [] [h] [] [];
      possibly_new_prop p2 false [] [h] [] [];
      possibly_new_prop p1 true [] [h] [] [];
      possibly_new_prop p2 true [] [h] [] [];
      add_search_clause [-(new_atomh h); -(new_literal p1); new_literal p2] [-(new_atomh h)];
      add_search_clause [-(new_atomh h); new_literal p1; -(new_literal p2)] [-(new_atomh h)]
  | Ap(TpAp(TmH(k),Set),p1) when k = !expoly -> process_pos_prop_3 h (Ap(TmH(!exSet),p1))
  | Ap(TmH(k),Lam(_,p1)) when k = !exSet ->
      let k = tm_id (Ap(TpAp(Prim(0),Set),Lam(Set,p1))) in
      Hashtbl.add eigen k h;
      if !using_guidance then
        begin
          let (sch,il) = tm_schema_and_instnames p in
          let schid = schema_id sch in
          let ct = Eigen(schid) in
          Hashtbl.add condstable (ct,k::il) ();
          apply_forward_guidance_rules ct (k::il)
        end;
      let c = TmH(k) in
      let p1c = tmsubst p1 0 c in
      if !eigensasinsts then possibly_new_inst c k;
      possibly_new_prop p1c true [] [h] [] [];
      add_search_clause [-(new_atomh h); new_literal p1c] [-(new_atomh h)]
  | Ap(TmH(k),p1) when k = !exSet ->
      let k = tm_id (Ap(TpAp(Prim(0),Set),p1)) in
      let c = TmH(k) in
      let p1c = Ap(p1,c) in
      if !eigensasinsts then possibly_new_inst c k;
      possibly_new_prop p1c true [] [h] [] [];
      add_search_clause [-(new_atomh h); new_literal p1c] [-(new_atomh h)]
  | _ ->
      let i = -(new_atomh h) in
      let (k,s) = tm_head_args p in
      Hashtbl.add posatmh k (i,s,h);
      List.iter
	(fun (j,s2,nh) ->
          if (not (i = -j) || !self_matings) then
	    begin
	      let jl = ref [] in
              List.iter2
		(fun m1 m2 ->
                     let d1 = Ap(Ap(TmH(!eqSet),m1),m2) in
		     jl := -(new_atom d1)::!jl;
		     possibly_new_prop d1 false [] [h] [nh] [])
		s s2;
              add_search_clause (i::j::!jl) [i; j]
            end)
	(Hashtbl.find_all negatmh k);
      process_pred_lin_shallow h i p true

let rec process_neg_prop q pr h p =
  if !verbosity > 4 then (Printf.printf "Maybe Processing neg %s %s\n" h (s_tm_to_str p); flush stdout);
  if q = 0 || q = 3 then (** don't use the prop model to delay very high priority or very low priority options **)
    process_neg_prop_2 h p  
  else
    let i = new_atomh h in
    let v = minisat_modelValue (-i) in
    if !verbosity > 9 then (Printf.printf "modelValue %d\n" v; flush stdout);
    if v > 0 then (** delay working on it if it's not on the branch identified by the current minisat assignment **)
      if q = 1 then
        if pr < 256 then
          PQueue1.insertWithPriority (pr*2+1) (ProcessNegProp(h,p))
        else
          PQueue2.insertWithPriority pr (ProcessNegProp(h,p))
      else if pr < 2048 then
        PQueue2.insertWithPriority (pr*2+1) (ProcessNegProp(h,p))
      else
        PQueue3.insertWithPriority pr (ProcessNegProp(h,p))
    else
      process_neg_prop_2 h p  
and process_neg_prop_2 h p =
  if !verbosity > 4 then Printf.printf "Processing neg %s %s\n" h (s_tm_to_str p);
  if !verbosity > 9 then Printf.printf "(PROCESSNEG \"%s\" %d %d)\n" h !abstract_time !middle_abstract_time;
  incr_abstract_time();
  process_neg_prop_3 h p
and process_neg_prop_3 h p =
  Hashtbl.add negpropdoneh h ();
  if !using_guidance then
    begin
      let (sch,il) = tm_schema_and_instnames p in
      let schid = schema_id sch in
      let ct = ProcessedNeg(schid) in
      Hashtbl.add condstable (ct,il) ();
      apply_forward_guidance_rules ct il;
      List.iter
        (fun q ->
          possibly_new_pos_prop q [] [] [h] [];
          possibly_new_neg_prop q [] [] [h] [])
        !guidance_cuts;
      List.iter
        (fun m -> possibly_new_inst m h)
        !guidance_insttrms;
      guidance_cuts := [];
      guidance_insttrms := [];
    end;
  match p with
  | Imp(p1,p2) ->
      possibly_new_prop p1 true [] [] [h] [];
      possibly_new_prop p2 false [] [] [h] [];
      add_search_clause [new_atomh h; new_literal p1] [new_atomh h];
      add_search_clause [new_atomh h; - (new_literal p2)] [new_atomh h]
  | All(Set,p1) ->
      let k = tm_id (Ap(TpAp(Prim(0),Set),Lam(Set,Ap(TmH(!neg),p1)))) in
      Hashtbl.add eigen k h;
      if !using_guidance then
        begin
          let (sch,il) = tm_schema_and_instnames p in
          let schid = schema_id sch in
          let ct = Eigen(schid) in
          Hashtbl.add condstable (ct,k::il) ();
          apply_forward_guidance_rules ct (k::il)
        end;
      let c = TmH(k) in
      let p1c = tmsubst p1 0 c in
      if !eigensasinsts then possibly_new_inst c k;
      possibly_new_prop p1c false [] [] [h] [];
      add_search_clause [new_atomh h; - (new_literal p1c)] [new_atomh h]
  | Ap(Ap(TpAp(TmH(k),Set),m1),m2) when k = !eqPoly -> process_neg_prop_3 h (Ap(Ap(TmH(!eqSet),m1),m2))
  | Ap(Ap(TmH(k),m1),m2) when k = !eqSet && m1 = m2 ->
      let i = new_atomh h in
      add_search_clause [i] [i]
  | Ap(Ap(TmH(k),m1),m2) when k = !eqSet ->
      let i = new_atomh h in
      if !enablecongclos then
	begin
	  if !verbosity > 19 then Printf.printf "trying cong clos !=\n";
	  List.iter
	    (fun (j,n1,n2,eh) ->
	      if not (Hashtbl.mem congeqnsh j) then (** already part of the cc **)
		begin (** otherwise check if the eqn is true on every remaining open branch, by asking minisat if adding j lit (-of eqn) is unsat **)
		  if not (minisat_search1 j) then
		    begin
		      Hashtbl.add congeqnsh j ();
		      congeqnlits := j::!congeqnlits;
		      congeqnhs := eh::!congeqnhs;
		      let (ccv,ccp,ccb,ccf) = cc in
		      extend_cong_graph cc n1;
		      extend_cong_graph cc n2;
		      cong_merge (ccp,ccb,ccf) n1 n2
		    end
		end)
	    !poseqnlist;
	  if !verbosity > 19 then Printf.printf "testing cong clos for %d: %s =? %s\n" i (s_tm_to_str m1) (s_tm_to_str m2);
	  if cong m1 m2 then
	    begin
	      Hashtbl.add congdeqnsh i ();
	      if !verbosity > 19 then Printf.printf "congclos gave clause for %d\n" i;
	      add_search_clause (i::!congeqnlits) (i::!congeqnlits) (** consider them all to be principal **)
	    end;
	end;
      possibly_new_inst m1 h;
      possibly_new_inst m2 h;
      let (k1,s1) = tm_head_args m1 in
      let (k2,s2) = tm_head_args m2 in
      if k1 = k2 then (** decomposition **)
	begin
          let jl = ref [] in
          List.iter2
            (fun n1 n2 ->
              let d1 = Ap(Ap(TmH(!eqSet),n1),n2) in
              jl := -(new_atom d1)::!jl;
              possibly_new_prop d1 false [] [] [h] [])
            s1 s2;
          add_search_clause (i::!jl) [i]
	end;
      if !eqn_sym_rule then
        begin
          let q = Ap(Ap(TmH(!eqSet),m2),m1) in
          let qh = tm_id q in
          let j = - (new_atomh qh) in
          possibly_new_prop q true [] [] [h] [];
          add_search_clause [i;j] [i];
        end;
      List.iter
	(fun (j,n1,n2,nh) ->
          if (not (i = -j) || !self_confrontations) then
            begin
              let d1 = Ap(Ap(TmH(!eqSet),m1),n1) in
              let d2 = Ap(Ap(TmH(!eqSet),m1),n2) in
              let d3 = Ap(Ap(TmH(!eqSet),m2),n1) in
              let d4 = Ap(Ap(TmH(!eqSet),m2),n2) in
              possibly_new_prop d1 false [] [nh] [h] [];
              possibly_new_prop d2 false [] [nh] [h] [];
              possibly_new_prop d3 false [] [nh] [h] [];
              possibly_new_prop d4 false [] [nh] [h] [];
              add_search_clause [i; j; -(new_atom d1); -(new_atom d3)] [i; j];
              add_search_clause [i; j; -(new_atom d1); -(new_atom d4)] [i; j];
              add_search_clause [i; j; -(new_atom d2); -(new_atom d3)] [i; j];
              add_search_clause [i; j; -(new_atom d2); -(new_atom d4)] [i; j]
            end)
	!poseqnlist;
      negeqnlist := (i,m1,m2,h)::!negeqnlist;
      process_pred_lin_shallow h i p false;
      process_func_lin_shallow h i m1 m2 false;
      process_func_lin_shallow h i m2 m1 false
  | TmH(k) when k = !fal -> ()
  | TmH(k) when k = !tru -> add_search_clause [(new_atomh h)] [(new_atomh h)]
  | Ap(Ap(TmH(k),p1),p2) when k = !conj ->
      possibly_new_prop p1 false [] [] [h] [];
      possibly_new_prop p2 false [] [] [h] [];
      add_search_clause [new_atomh h; - (new_literal p1); - (new_literal p2)] [new_atomh h]
  | Ap(Ap(TmH(k),p1),p2) when k = !disj ->
      possibly_new_prop p1 false [] [] [h] [];
      possibly_new_prop p2 false [] [] [h] [];
      add_search_clause [new_atomh h; -(new_literal p1)] [new_atomh h];
      add_search_clause [new_atomh h; -(new_literal p2)] [new_atomh h]
  | Ap(Ap(TmH(k),p1),p2) when k = !iff ->
      possibly_new_prop p1 false [] [] [h] [];
      possibly_new_prop p2 false [] [] [h] [];
      possibly_new_prop p1 true [] [] [h] [];
      possibly_new_prop p2 true [] [] [h] [];
      add_search_clause [new_atomh h; new_literal p1; new_literal p2] [new_atomh h];
      add_search_clause [new_atomh h; - (new_literal p1); - (new_literal p2)] [new_atomh h]
  | Ap(TpAp(TmH(k),Set),p1) when k = !expoly -> process_neg_prop_3 h (Ap(TmH(!exSet),p1))
  | Ap(TmH(k),Lam(_,p1)) when k = !exSet ->
    let i = new_atomh h in
    List.iter
      (fun (m,ih) ->
        let p1m = tmsubst p1 0 m in
        Hashtbl.add clause_inst [i; -new_literal p1m] m;
        possibly_new_prop p1m false [] [] [h] [ih];
        add_search_clause [i; -new_literal p1m] [i])
      !instlist;
    if !dynamic_shallowrules then
      begin
	let srules = prop_as_shallow_rules (-i) h p false in
	if not (srules = []) then
	  begin
	    if !verbosity > 25 then (Printf.printf "dynamic srules %d %s\n" (List.length srules) h; flush stdout);
	    Hashtbl.add shallowatom i ();
	    List.iter
	      (fun (hss,srule) ->
		Hashtbl.add shallow_src_neg hss h;
		register_shallow_rule (hss,srule))
	      srules
	  end
      end;
    alllist := (i,false,p1,h) :: !alllist
  | Ap(TmH(k),p1) when k = !exSet ->
    let i = new_atomh h in
    List.iter
      (fun (m,ih) ->
        let p1m = Ap(p1,m) in
        Hashtbl.add clause_inst [i; -new_literal p1m] m;
        possibly_new_prop p1m false [] [] [h] [ih];
        add_search_clause [i; -new_literal p1m] [i])
      !instlist;
    alllist := (i,false,p1,h) :: !alllist
  | _ ->
    let i = new_atomh h in
    let (k,s) = tm_head_args p in
    Hashtbl.add negatmh k (i,s,h);
    List.iter
      (fun (j,s2,nh) ->
	let jl = ref [] in
        List.iter2
          (fun m1 m2 ->
            if (not (i = -j) || !self_matings) then
              begin
                let d1 = Ap(Ap(TmH(!eqSet),m1),m2) in
	        jl := -(new_atom d1)::!jl;
	        possibly_new_prop d1 false [] [nh] [h] []
              end)
          s2 s;
        add_search_clause (j::i::!jl) [j; i])
      (Hashtbl.find_all posatmh k);
    process_pred_lin_shallow h i p false

let process_inst h m =
  if !verbosity > 4 then Printf.printf "Processing inst %s %s\n" h (s_tm_to_str m);
  if !verbosity > 9 then Printf.printf "(PROCESSINST \"%s\" %d %d)\n" h !abstract_time !middle_abstract_time;
  incr_abstract_time();
  Hashtbl.add instdoneh h ();
  if !using_guidance then
    begin
      let (sch,il) = tm_schema_and_instnames m in
      let schid = schema_id sch in
      let ct = ProcessedInst(schid) in
      Hashtbl.add condstable (ct,il) ();
      apply_forward_guidance_rules ct il;
      List.iter
        (fun q ->
          possibly_new_pos_prop q [] [] [] [h];
          possibly_new_neg_prop q [] [] [] [h])
        !guidance_cuts;
      List.iter
        (fun m -> possibly_new_inst m h)
        !guidance_insttrms;
      guidance_cuts := [];
      guidance_insttrms := [];
    end;
  List.iter
    (fun (i,b,p1,ph) ->
      let p1m = tmsubst p1 0 m in
      let l = new_literal p1m in
      let j = if b then l else -l in
      Hashtbl.add clause_inst [i; j] m;
      possibly_new_prop p1m b [] (if b then [ph] else []) (if b then [] else [ph]) [h];
      add_search_clause [i; j] [i])
    !alllist;
  instlist := (m,h) :: !instlist

let default_inst_option () =
  let m = Ap(TpAp(Prim(0),Set),TmH(!fal)) in
  let h = tm_id m in (** here the term and the 'schema' are the same since m only has logical constants (epsilon and false) **)
  if Hashtbl.mem pos_watch_pos h then Printf.printf "(WATCH POS INST \"%s\" %d)\n" h !middle_abstract_time;
  if Hashtbl.mem neg_watch_pos h then Printf.printf "(WATCH NEG INST \"%s\" %d)\n" h !middle_abstract_time;
  if Hashtbl.mem pos_rote_inst h || (!using_guidance && pos_guidance_inst m) then
    begin
      if !verbosity > 19 then Printf.printf "priority 0 0 due to pos rote inst\n";
      PQueue0.insertWithPriority 0 (ProcessInst(h,m))
    end
  else
    possibly_new_inst_priority m 1000 (** otherwise low priority, here mainly as a fallback **)

let reset_guidance () =
  Hashtbl.clear cutproductionrule;
  Hashtbl.clear insttrmproductionrule;
  Hashtbl.clear goodposproductionrule;
  Hashtbl.clear badposproductionrule;
  Hashtbl.clear goodnegproductionrule;
  Hashtbl.clear badnegproductionrule;
  Hashtbl.clear goodinstproductionrule;
  Hashtbl.clear badinstproductionrule;
  Hashtbl.clear condstable;
  guidance_cuts := [];
  guidance_insttrms := []

let reset_atoms () =
  atom_count := 0;
  Hashtbl.clear atomh;
  Hashtbl.clear atomh_rev;
  Hashtbl.clear atomh_prop_rev;
  Hashtbl.clear atomh_sch_rev;
  Hashtbl.clear shallowatom

let reset_clauses () =
  clauses := []

let reset_search () =
  PQueue0.reset();
  PQueue1.reset();
  PQueue2.reset();
  PQueue3.reset();
  instlist := [];
  alllist := [];
  Hashtbl.clear posatmh;
  Hashtbl.clear negatmh;
  poseqnlist := [];
  negeqnlist := [];
  Hashtbl.clear atmshlinh;
  Hashtbl.clear trmshlinh;
  Hashtbl.clear posproph;
  Hashtbl.clear negproph;
  Hashtbl.clear insth;
  Hashtbl.clear pospropdoneh;
  Hashtbl.clear negpropdoneh;
  Hashtbl.clear instdoneh;
  noinsts := true;
  Hashtbl.clear ordered_salted_h;
  Hashtbl.clear funcdef;
  Hashtbl.clear preddef;
  let (ccv,ccp,ccb,ccf) = cc in
  Hashtbl.clear ccv;
  Hashtbl.clear ccp;
  Hashtbl.clear ccb;
  Hashtbl.clear ccf;
  Hashtbl.clear congeqnsh;
  congeqnlits := [];
  congeqnhs := [];
  Hashtbl.clear congdeqnsh;
  middle_abstract_time := 0;
  middle_search := true

let search_main () =
  if !noinsts then default_inst_option();
  while true do
    try
      let fallback3 () = let (pr,so) = PQueue3.getNext() in (pr,so,3) in
      let fallback2 () = try let (pr,so) = PQueue2.getNext() in (pr,so,2) with Not_found -> fallback3() in
      let fallback1 () = try let (pr,so) = PQueue1.getNext() in (pr,so,1) with Not_found -> fallback2() in
      let fallback0 () = try let (pr,so) = PQueue0.getNext() in (pr,so,0) with Not_found -> fallback1() in
      let (pr,so,q) =
        try
          if !abstract_time mod 10 = 0 then
	    if !abstract_time mod 100 = 0 then
               if !abstract_time mod 1000 = 0 then
                 fallback3()
               else
                 fallback2()
            else
              fallback1()
          else
            fallback0()
        with Not_found -> fallback0()
      in
      if !verbosity > 28 then Printf.printf "took option from queue %d abstime %d middleabstime %d\n" q !abstract_time !middle_abstract_time;
      match so with
      | ProcessPosProp(h,p) -> if not (Hashtbl.mem pospropdoneh h) then process_pos_prop q pr h p
      | ProcessNegProp(h,p) -> if not (Hashtbl.mem negpropdoneh h) then process_neg_prop q pr h p
      | ProcessInst(h,m) -> if not (Hashtbl.mem instdoneh h) then process_inst h m
    with Not_found -> raise NoOption
  done

